<?php
/**
 * @file
 * Functions related to the Garmin Connector Catalogue
 */

/**
 * Rebuild the catalog.
 *
 * Sometimes the files on the filesystem change. This could happen because
 * a particular activity has been purged or when activities are added
 *
 * @param bool $debug
 *   Print interesting information
 */
function garmin_connector_rebuild_catalogue($debug = NULL) {
  global $_garmin_connector_from_drush;
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  if ($_garmin_connector_from_drush == 1) {
    $debug = fopen('php://stdout', 'w');
  }
  elseif ($debug != NULL) {
  }
  else {
    $catalog_debug_destination = "public://garmin_connector/debug-catalogue.txt";
    $cpath = drupal_realpath($catalog_debug_destination);
    $debug = fopen($cpath, "w");
  }
  // Let's start with finding all files.
  $path = drupal_realpath(garmin_connector_get_config('dir_gc'));
  fwrite($debug, t("@fn catalogue path is (@path)\n",
    array('@path' => $path, '@fn' => $fn)
  ));
  $allfiles = file_scan_directory($path, "/tcx|json/");
  $all_ids = array();
  foreach ($allfiles as $filename => $value) {
    $count1 = preg_match_all("/(\d+)\.(json|tcx)$/", $filename, $matches);
    if ($count1 > 0) {
      $id = $matches[1][0];
      $count2 = preg_match("/(cycling|running|swimming|other|fitness_equipment)/", $filename, $matches2);
      if ($count2 > 0) {
        $sport = $matches2[0];
        $all_ids[$id] = array('id' => intval($id), 'sport' => $sport);
      }
      else {
        // Didn't see a sport in the filename $filename.
      }
    }
    else {
      // The file with filename $filename is not json or txt with an ID.
    }
  }
  fwrite($debug,
      t("@fn There are this many ids in the catalogue: (@count)\n",
      array(
        '@fn' => $fn,
        '@count' => count($all_ids))));
  $activity_summaries = array();
  foreach ($all_ids as $id => $contents) {
    // Try to get the overview.
    $dest = garmin_connector_get_config(
      'dir_summary_' . $contents['sport'])
      . "/" . $contents['id'] . ".json";
    $path = drupal_realpath($dest);
    if (is_file($path)) {
      $contents = file_get_contents($path);
      $activity = json_decode($contents);
      if (!is_object($activity)) {
        $message = t("@fn Failed to decode json for the file at @path\n",
          array(
            '@fn' => $fn,
            '@path' => $path,
          )
        );
        fwrite($debug, $message);
        watchdog("garmin_connector",
          "@fn Failed to decode json for the file at @path\n",
          array(
            '@fn' => $fn,
            '@path' => $path,
          )
        );
        return;
      }
      $activity_summaries['by-id'][$activity->activityId] = $activity;
      $time = new DateTime($activity->activitySummary->BeginTimestamp->value);
      $better_timestamp = $time->format('Y-m-d H:i:s');
      $activity_summaries['by-date'][$better_timestamp] = $activity;
    }
    else {
      fwrite($debug,
        t("@fn Failed to find summary here (@path). This is unexpected.\n",
        array(
          '@fn' => $fn,
          '@path' => $path)
      ));
      return;
    }
  }
  fwrite($debug, t("Done parsing items in the catalogue\n"));
  $destination_location = garmin_connector_get_config('gc_catalogue');
  $catpath = drupal_realpath($destination_location);
  arsort($activity_summaries['by-id']);
  $result1 = file_put_contents($catpath,
  garmin_connector_get_json_string($activity_summaries['by-id']));
  if ($result1 === FALSE) {
    fwrite($debug, t("There was an error writing the system catalogue to the path @catpath\n",
      array('@catpath' => $catpath)));
  }
  else {
    fwrite($debug,
      t("@fn Wrote system catalog to here (@catpath)\n",
      array(
        '@catpath' => $catpath,
        '@fn' => $fn)));
  }
  // Now write by date, to make sure.
  arsort($activity_summaries['by-date']);
  $datepath = drupal_realpath(garmin_connector_get_config('gc_catalogue_by_date'));
  $putter = garmin_connector_get_json_string($activity_summaries['by-date']);
  $result2 = file_put_contents($datepath, $putter);
  if ($result2 === FALSE) {
    fwrite($debug, t("There was an error writing the system catalogue by date to the path !datepath\n",
      array(
        '!datepath' => var_export($datepath, TRUE),
      )
    ));
  }
  fwrite($debug, t("All done rebuilding catalogue. Results of writing catalgue is @cat and results of writing catalogue by date is @cat2\n",
    array(
      '@cat' => var_export($result1, TRUE),
      '@cat2' => var_export($result2, TRUE),
    )
  ));
}

/**
 * Change the key in array to be not zero-indexed but based on the ActivityId.
 *
 * @param array $activities
 *   An array containing a list of zero-indexed workouts
 *
 * @return array
 *   An array containing the same workouts but indexed on their indexId
 */
function garmin_connector_key_by_id($activities) {
  $returner = array();
  if (count($activities) == 0) {
    // Nothing to do here.
    return;
  }
  foreach ($activities as $activity) {
    $returner[$activity->activityId] = $activity;
  }
  return $returner;
}

/**
 * Install sample data.
 *
 * @param filehandle $debug
 *   A filehandle to write information
 *
 * @throws
 *   - Errors when not logged in. Because this method alters the filesystem
 *   it requires authentication. The method itself needs to be callable for demo
 *   purposes.
 *   - Errors in network communication will also be reported.
 */
function garmin_connector_install_sample_data($debug = NULL) {
  global $_garmin_connector_from_drush;
  if ($_garmin_connector_from_drush == 1) {
  }
  elseif (!user_access("manage activities") == TRUE) {
    throw new Exception(t("You need to be given the permission 'Manage Activities' to download sample data"));
  }
  $url_base = "http://ironman.teamchad.ca/sites/ironman.teamchad.ca/";
  $url_base .= "files/garmin_connector/garmin-connector-sample-data/";
  $files = array(
    // An easy run.
    '/activities/summary/running/403540379.json',
    '/activities/details/running/403540379.json',
    '/activities/tcx/running/403540379.tcx',
    '/activities/plot-data/403540379.json',
    // Easy run.
    'activities/summary/running/401514442.json',
    'activities/details/running/401514442.json',
    'activities/tcx/running/401514442.tcx',
    'activities/plot-data/401514442.json',
    // Zone 2 bike.
    'activities/details/cycling/402599290.json',
    'activities/summary/cycling/402599290.json',
    'activities/tcx/cycling/402599290.tcx',
    'activities/plot-data/402599290.json',
    // Bike with intervals.
    'activities/summary/cycling/401954572.json',
    'activities/details/cycling/401954572.json',
    'activities/tcx/cycling/401954572.tcx',
    'activities/plot-data/401954572.json',
    // Ironman swim.
    'activities/summary/swimming/349919979.json',
    'activities/details/swimming/349919979.json',
    'activities/tcx/swimming/349919979.tcx',
    'activities/plot-data/349919979.json',
    // Ironman bike.
    'activities/summary/cycling/349920054.json',
    'activities/details/cycling/349920054.json',
    'activities/tcx/cycling/349920054.tcx',
    'activities/plot-data/349920054.json',
    // Ironman run.
    'activities/summary/running/349920083.json',
    'activities/details/running/349920083.json',
    'activities/tcx/running/349920083.tcx',
    'activities/plot-data/349920083.json',
  );
  $refresh = "no";
  foreach ($files as $file) {
    $url = $url_base . $file;
    $destination = "public://garmin_connector/" . $file;
    // Get the file.
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
    if ($refresh == "no" && file_exists($destination)) {
      $message = t("Not getting sample data for @file because it alrady exists.",
      array('@file' => $file));
      $debug && fwrite($debug, $message . "\n");
    }
    else {
      $message = t("Getting sample data: !url",
        array('!url' => $url)
      );
      watchdog("garmin_connector", $message);
      $debug && fwrite($debug, $message . "\n");
      $result = curl_exec($ch);
      if (curl_error($ch)) {
        throw new Exception(t("An error occurred while trying to get sample data from the URL !url. The error was @error.",
        array(
          '!url' => "<a href='" . $url . "'>$url</a>",
          '@error' => curl_error($ch),
        )
        ));
      }
      file_put_contents($destination, $result);
    }
  }
  $debug && fwrite($debug, t("Done. Now inserting sample workout sets"));
  $select = db_insert("garmin_connector_workout_sets")
    ->fields(array(
      'user'     => "all",
      'set_name' => "sample-running",
      'gc_id'    => "401514442",
      'gc_name'  => "Easy Run",
      'gc_type'  => "running",));
  $select->execute();
  $select = db_insert("garmin_connector_workout_sets")
    ->fields(array(
      'user'     => "all",
      'set_name' => "sample-running",
      'gc_id'    => "403540379",
      'gc_name'  => "Easy Run",
      'gc_type'  => "running",));
  $select->execute();
  garmin_connector_rebuild_catalogue();
}

/**
 * Provide a list of workouts that are available in the local catalogue.
 *
 * When I first wrote this software I spoke REST to Garmin Connect a lot.
 * This made performance
 * **TERRIBLE** so the only option was to cache everything locally
 *
 * @param string $by_date
 *   Give the list sorted by date
 *
 * @return object
 *   The activities in the catalogue
 *
 * @throws
 *   A file containing activities must have been built by
 *   garmin_connector_rebuild_catalogue() before this can be called.
 *
 * @see garmin_connector_rebuild_catalogue()
 */
function garmin_connector_get_activities_from_local($by_date = NULL) {
  if ($by_date == NULL) {
    $destination = garmin_connector_get_config('gc_catalogue');
  }
  else {
    $destination = garmin_connector_get_config('gc_catalogue_by_date');
  }
  if (array_key_exists("activities_from_file-$destination", $GLOBALS)) {
    // "These activities have already been read from the file.
    // No need to load them again");
    return $GLOBALS["activities_from_file-$destination"];
  }
  $realpath = drupal_realpath($destination);
  if (!file_exists($realpath)) {
    // This may happen in a fresh install. This needs to be an exception.
    // Because without data, many things will behave unexpectedly.
    // The Exception provides the best reliability.
    throw new Exception(t("The file containing activities (@file) does not exist. This can happen when there is no data or when the Garmin Connector system catalogue has never been built. Go to the Garmin Connector catalogue maintenance page !here and add some data! You can also install sample data with drush using the command garmin-connector-sample-data.",
    array(
      '@file' => $realpath,
      '!here' => "<a href='" .
      url("garmin-connector/activity-manager.html") . "'>" . t("here") . "</a>",
    )));
  }
  $catalogue_pulled = FALSE;
  $apc_exists = FALSE;
  if (function_exists('apc_store')) {
    // watchdog("garmin_connector","APC store exists.");
    $apc_exists = TRUE;
  }
  // If APC is present, try to use it.
  if ($apc_exists === TRUE) { 
    // watchdog("garmin_connector","Checking APC cache");
    // If it's in the cache, pull it
    $act2 = apc_fetch("garmin_connector_catalogue");
    if (!$act2 === FALSE) {
      // Nothing more to see here. Move along.
      watchdog("garmin_connector","Catalogue pulled from cache. It's really good that you have APC enabled!");
      $catalogue_pulled = TRUE;
    }
  }
  if ($catalogue_pulled === FALSE) {
    $contents = file_get_contents($realpath);
    $act2 = json_decode($contents);
    if ($apc_exists === TRUE) {
      apc_store('garmin_connector_catalogue', $act2, 0);
      // watchdog("garmin_connector","Stored the catalogue in APC.");
    }
  }
  $GLOBALS["activities_from_file-$destination"] = $act2;
  return $GLOBALS["activities_from_file-$destination"];
}

/**
 * Write the activities.
 *
 * @param object $activities
 *   A list of activities
 */
function garmin_connector_write_activities($activities) {
  $destination = garmin_connector_get_config('gc_catalogue');
  $realpath = drupal_realpath($destination);
  file_put_contents($realpath,
    garmin_connector_get_json_string($activities));
}

/**
 * A debugging routine.
 *
 * Intended to be used within Drush
 * drush -r /var/lib/drupal7 -l http://yoursite php-eval \
 * 'garmin_connector_make_json_from_files("filename1","filename2")'
 *
 * @param string $details_filename
 *   The filename for the details file
 *
 * @param filehandle $debug
 *   A filehandle to write output
 */
function garmin_connector_make_json_from_details_file($details_filename, $debug) {
  // This is for debugging. Use it within drush.
  $overview_filename
    = preg_replace("/details/", "summary", $details_filename);
  fwrite($debug, t("overview_filename is !overview_filename and details_filename is !details_filename\n",
    array(
      '!overview_filename' => $overview_filename,
      '!details_filename'  => $details_filename,
    )
  ));
  garmin_connector_make_json_from_files($overview_filename, $details_filename,
  $debug);
}

/**
 * Intended to be invoked from garmin_connector_make_json_from_details_file.
 *
 * and thus executed from drush
 * drush -r /var/lib/drupal7 -l http://yoursite php-eval 
 * 'garmin_connector_make_json_from_files()'
 */
function garmin_connector_make_json_from_files($overview_filename,
$details_filename, $debug) {
  if (!is_file($details_filename)) {
    fwrite($debug,
      t("The details filename named !details_filename is not a file\n",
      array(
        '!details_filename' => $details_filename,
      )));
    return;
  }
  if (!is_file($overview_filename)) {
    fwrite($debug,
      t("The overview file (!overview_filename) is not a file\n",
      array(
        '!overview_filename' => $overview_filename,
      )
      ));
    return;
  }
  // Call the real method.
  fwrite($debug,
    t("Getting overview data from the file !overview_filename\n",
    array(
      '!overview_filename' => $overview_filename,
    )));
  $overview_json = file_get_contents($overview_filename);
  $overview = json_decode($overview_json);
  fwrite($debug,
    t("Getting details from the file named !details_filename\n",
    array(
      '!details_filename' => $details_filename,
    )));
  $details_json = file_get_contents($details_filename);
  $details = json_decode($details_json);
  $details->filename = $details_filename;
  garmin_connector_make_json_from_objects($overview, $details, $debug);
}

/**
 * Calculate a whole bunch of performance metrics intended to be plotted in GC.
 *
 * @param object $details
 *   An object representing the details of a Garmin Connect workout.
 *
 * @param filehandle $debug
 *   A filehandle to write information.
 *
 * @return array
 *   A set of points intended go be read by ajax 
 *   and plotted into a browser window
 */
function garmin_connector_get_measurements($details, $debug) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  fwrite($debug,
    t("@fn -- Parsing measurements\n",
    array(
      '@fn' => $fn,
    )
  ));
  if (!is_object($details)) {
    fwrite($debug,
      "@fn -- Something is very wrong, details is not an object.\n",
      array(
        '@fn' => $fn,
      )
    );
    drupal_set_message(
      t("Details is not an object. This is very bad."),
      'error');
    return;
  }
  else {
    // wtf?
    // $returned = var_export($details,TRUE);
    // fwrite($_garmin_connector_debug, "$fn -- export: $returned\n");
  }
  $index = array();
  foreach (array("directHeartRate",
      "sumDuration",
      "directPower",
      "directPace") as $thingy) {
    if (property_exists($details->activity, "measurements")) {
      foreach ($details->activity->measurements as $key => $metric) {
        $num = $metric->metricsIndex;
        $name = $metric->key;
        if ($name == $thingy) {
          $index[$name] = array('units-key' => $key, 'metrics-key' => $num);
        }
      }
    }
    else {
      fwrite($debug,
        t("@fn -- A measurement type @thingy not found for @activityname with type @activitytype\n",
        array(
          '@fn' => $fn,
          '@thingy' => $thingy,
          '@activityname' => $details->activity->activityName,
          '@activitytype' => garmin_connector_get_type($details->activity),
        )
      ));
    }
  }
  $count = 0;
  $points = array();
  $points['heartrate'] = array();
  $points['power'] = array();
  if (array_key_exists('directHeartRate', $index)) {
    $units = $details->activity
      ->measurements[$index['directHeartRate']['units-key']]->unitDisplay;
    $points['heartrate']['units'] = $units;
  }
  if (array_key_exists('directPower', $index)) {
    $units = $details->activity
      ->measurements[$index['directPower']['units-key']]->unitDisplay;
    $points['power']['units'] = $units;
  }
  if (array_key_exists('directPace', $index)) {
    $units = $details->activity->measurements[$index['directPace']['units-key']]
      ->unitDisplay;
    $points['pace']['units'] = $units;
  }
  $base = $details->activity;
  if (property_exists($base, "metrics")) {
    foreach ($base->metrics as $lap => $thislap) {
      if (!isset($lap)) {
      }
      $count++;
      $metrics = $thislap->metrics;
      $time = $metrics[$index['sumDuration']['metrics-key']];
      if (array_key_exists('directHeartRate', $index)) {
        $points['heartrate']['data'][$time] = array(
          $time,
          $metrics[$index['directHeartRate']['metrics-key']]);
      }
      if (array_key_exists('directPower', $index)) {
        $points['power']['data'][$time] = array(
          $time,
          $metrics[$index['directPower']['metrics-key']]);
      }
      if (array_key_exists('directPace', $index)) {
        // Probably not meters per second.
        $measurement = $metrics[$index['directPace']['metrics-key']];
        if ($measurement == 0) {
          $points['pace'][$time] = array($time, 0);
        }
        else {
          $points['pace']['data'][$time] = array($time, $measurement);
        }
      }
    }
  }
  return $points;
}

/**
 * Given a set of activities, return activities between certain dates.
 *
 * @param DateTime $start_date
 *   The start date. Optional.
 *
 * @param DateTime $end_date
 *   The end date. Optional.
 *
 * @param int $limit
 *   Return a maximum of this many activities.
 *
 * @return array
 *   An array of activities, keyed by sport
 */
function garmin_connector_get_activities($start_date = NULL, $end_date = NULL, $limit = NULL) {
  $returner = array();
  foreach (garmin_connector_get_sports() as $sport) {
    $stuff = garmin_connector_get_activities_by_sport($sport,
      $start_date,
      $end_date,
      NULL);
    $returner[$sport] = $stuff;
  }
  return $returner;
}

/**
 * Get activities by sport.
 *
 * This works with file that have been downloaded to the filesystem.
 *
 * Note that to download new activities, you must use 
 * garmin_connector_refresh_activities.
 * This makes sense because it takes some time to do this.
 *
 * What I normally do is this:
 * 1. Run
 * 2. My 910xt uploads the data to Garmin Connect
 * 3. I name the activity
 * 4. I run drush
 * drush -r /var/lib/drupal7 -l yoursite garmin-connector-refresh-activities
 * done!
 *
 * See the readme for what to do when you've changed a workout in Garmin Connect
 *
 * @param string $sport
 *   The sport name, as found in Garmin Connect. Pretty much has to be
 *   in the list. defined in garmin_connector_get_sports
 *
 * @param DateTime $start_date
 *   The start date. Optional.
 *
 * @param DateTime $end_date
 *   The end date. Optional.
 *
 * @param int $limit
 *   Return a maximum of this many activities.
 *
 * @return array
 *   An array of activities, keyed by BeginDate of the activity
 */
function garmin_connector_get_activities_by_sport($sport, $start_date = NULL, $end_date = NULL, $limit = 10) {
  $dir = garmin_connector_get_config('dir_summary_' . $sport);
  $fp = opendir($dir);
  $list = array();
  $realdir = drupal_realpath($dir);
  while (FALSE !== ($file = readdir($fp))) {
    $thingy = "$dir/$file";
    $realpath = drupal_realpath($thingy);
    if (is_file($realpath) == FALSE) {
      continue;
    }
    if (is_link($realdir . "/" . $file)) {
      continue;
    }
    $time = filemtime($realpath);
    // Now, compare the dates.
    if ($start_date != NULL) {
      if (($time <= $end_date->getTimestamp() &&
        ($time >= $start_date->getTimestamp()))) {
        $list[$time] = $realpath;
      }
    }
  }
  asort($list);
  $pruned = array_slice($list, ($limit * -1), NULL, TRUE);
  $returner = array();
  foreach ($pruned as $timestamp => $filename) {
    if (!isset($timestamp)) {
    }
    // By doing this here we restrict it to $limit to open json_decode.
    $contents = file_get_contents($filename);
    $obj = json_decode($contents);
    $time = new DateTime($obj->activitySummary->BeginTimestamp->value);
    $better_timestamp = $time->format('Y-m-d H:i:s');
    $returner[$better_timestamp] = $obj;
  }
  krsort($returner);
  return $returner;
}

/**
 * Get a formatted date.
 *
 * DateTime does not always have the capability to parse useful strings.
 *
 * @param string $thing
 *   An english date.
 *
 * @return DateTime
 *   A DateTime object for the formatted date.
 */
function garmin_connector_get_date($thing) {
  $formatter = "Y-m-d";
  $date = new DateTime();
  switch ($thing) {
    case 'first day of last month':
      return $date->modify('first day of last month')->format($formatter);

    case 'last day of last month':
      return $date->modify('last day of last month')->format($formatter);

    case 'first day of this month':
      return $date->modify('first day of this month')->format($formatter);

    case 'last day of this month':
      return $date->modify('last day of this month')->format($formatter);

    case "three days ago":
      return $date->modify('today -3 day')->format($formatter);

    case "one week ago":
      return $date->modify('today -7 day')->format($formatter);

    case "two weeks ago":
      return $date->modify('today -14 day')->format($formatter);

    case "one month ago":
      return $date->modify('today -30 day')->format($formatter);

    case "today":
      return $date->format($formatter);
  }
  $dt = new DateTime($thing);
  return $dt->format($formatter);
}

/**
 * Get the last ten activities.
 *
 * @param object $a
 *   An object as defined by Garmin Connect
 *   This object should be a list of workouts keyed on a number *exactly* 
 *   as downloaded from Garmin Connect
 *   Look inside to see what the object should look like. It's obvious.
 *
 * @param string $wanttype
 *   The type of activity you want
 *
 * @param bool $include_graph
 *   Include the graph?
 *
 * @return array
 *   A drupal render array
 */
function garmin_connector_get_last_ten_activities($a, $wanttype = NULL, $include_graph = FALSE) {
  $returner['#prefix'] = "<div class='last-ten-activities last-ten last-ten-";
  $returner['#prefix'] .= ($wanttype ? $wanttype : "all") . " clearfix'>\n";
  $count = 0;
  $ignore_regex = variable_get('garmin_connector_key_workout_block_ignore_regex');
  foreach ($a as $key => $value) {
    if (!defined($key)) {
    }
    if ($count > 10) {
      continue;
    }
    $type = garmin_connector_get_type($value);
    if ($wanttype != NULL && $type != $wanttype) {
      continue;
    }
    $name = $value->activityName;
    if (strlen($ignore_regex) > 0) {
      $matches = preg_match("/$ignore_regex/i", $name);
      if ($matches > 0) {
        continue;
      }
      else {
      }
    }
    if (property_exists($value->activitySummary, "SumElapsedDuration")) {
      $duration = $value->activitySummary->SumElapsedDuration->display;
    }
    else {
      $duration = $value->activitySummary->SumDuration->display;
    }
    $count++;
    $date = $value->activitySummary->BeginTimestamp->display;
    $destination_json = garmin_connector_get_config('dir_plot_data') . "/";
    $destination_json .= $value->activityId . ".json";
    $json_url = file_create_url($destination_json);
    $this_html = "<a class='quick-view'\nhref=''\nonclick='garmin_connector_";
    $this_html .= "create_plot(\"$json_url\",\"#plot-container-0\"); ";
    $this_html .= "return false;'>Quick View</a>";
    $row = array(
      '#theme'    => 'garmin_connector_activity_row',
      '#date'     => $date,
      '#duration' => $duration,
      '#name'     => $name,
      '#gc-id'    => $value->activityId,
      '#classes'  => array(
        'last-ten-workout',
        'last-ten-two',
      ),
      '#children' => array(
        'quick-view-thing' => array(
          '#markup' => $this_html,
        ),
      ),
    );
    if ($include_graph == TRUE) {
      $row['#children'][] = garmin_connector_get_plot_container($count, $value->activityId);
    }
    $renderable = array(
      '#children' => array(
        '1' => $row,
      ),
      '#theme'      => 'garmin_connector_contextual_links_wrapper',
      '#gc-id'      => $value->activityId,
      '#gc-type'    => $type,
      '#gc-name'    => $name,
      '#gc-classes' => array(
        "last-ten-workout",
        "last-ten-workout-3",
      ),
      '#gc-css-id'  => 'last-ten-' . $value->activityId,
    );
    $returner["activity-$count"]  = $renderable;
  }
  $returner['#suffix'] = "</div><!-- /last-ten-activities -->";
  // Put the key at the bottom.
  $returner['key'] = garmin_connector_colors_get_key();
  return $returner;
}

/**
 * Get workout sets by ID.
 *
 * @param int $gc_id
 *   A garmin connect workout ID
 *
 * @return array
 *   An associative array containing the workout sets
 */
function garmin_connector_get_my_workout_sets($gc_id) {
  $names = db_select('garmin_connector_workout_sets', 's')
    ->fields("s");
  $or = db_or();
  $or->condition("s.gc_id", $gc_id, "=");
  $or->condition("s.gc_id", "all", "=");
  $names->condition($or);
  $handle = $names->execute();
  if (!$handle) {
    return array();
  }
  $options = array();
  $count = 0;
  while ($row = $handle->fetchAssoc()) {
    $count++;
    $options[$row['id_workout_sets']] = $row['set_name'];
  }
  return $options;
}

/**
 * Get the workouts in a workout set (by name).
 * 
 * @param string $name
 *   The name of a workout set
 *
 * @return array
 *   An associative array containing the workouts in a workout set.
 */
function garmin_connector_get_workout_set($name) {
  $in = array();
  $select = db_select('garmin_connector_workout_sets', 's')
    ->fields('s')
    ->condition('s.set_name', $name, '=')
    ->execute();
  if (!$select) {
    drupal_set_message(
      t("No workout sets were found with the name (@name)",
      array('@name' => $name)
    ), 'warning');
    return $in;
  }
  while ($row = $select->fetchAssoc()) {
    $in[$row['gc_id']] = $row;
  }
  return $in;
}


/**
 * Trigger the purge of Activities titled 'Untitled' for the past 30 days.
 *
 * Intended to be called from the web.
 */
function garmin_connector_trigger_purge_untitled($form, &$form_state) {
  garmin_connector_purge_untitled(NULL, NULL, NULL);
}

/**
 * Purge recent workouts that have the title 'Untitled'.
 *
 * Problem - how recent? I think a month.
 *
 * @param string $f
 *   The earlist date to search
 *
 * @param string $t
 *   The latest date to search
 *
 * @param object $debug
 *   A filehandle to print debug information
 *
 * @todo
 *   Throwing Exceptions would be useful.
 */
function garmin_connector_purge_untitled($f = NULL, $t = NULL, $debug = NULL) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  if (!isset($debug)) {
    $d1 = "public://garmin_connector/debug-purge-untitled.txt";
    $f1 = drupal_realpath($d1);
    $debug = fopen($f1, 'w');
  }
  if ($f == NULL) {
    $f = new DateTime(garmin_connector_get_date('one month ago'));
  }
  if ($t == NULL) {
    $t = new DateTime(garmin_connector_get_date('today'));
  }
  fwrite($debug,
    t("@fn Getting activities from (@f) to (@t)\n",
    array(
      '@fn' => $fn,
      '@f' => $f->format('Y-m-d H:i:s'),
      '@t' => $t->format('Y-m-d H:i:s'),
    ))
  );
  $activities = garmin_connector_get_activities_from_local();
  $selected = array();
  foreach ($activities as $activity) {
    $time = new DateTime($activity->activitySummary->BeginTimestamp->value);
    $title = $activity->activityName;
    if ($time < $f) {
      continue;
    }
    elseif ($time > $t) {
      continue;
    }
    elseif ($title == "Untitled") {
      array_push($selected, $activity->activityId);
    }
    else {
    }
  }
  if (count($selected) > 0) {
    garmin_connector_purge_activities($selected);
    garmin_connector_rebuild_catalogue();
  }
}

/**
 * Get the current list of workout sets.
 *
 * @param int $id
 *   That contain only this workout id
 *
 * @param bool $exclude
 *   That do not contain $id
 *
 * @return array
 *   An associative array containing the names of workout sets
 */
function garmin_connector_get_workout_set_names($id = NULL, $exclude = FALSE) {
  $query = db_select('garmin_connector_workout_sets', 's');
  $query->addField('s', 'gc_id', 'gc_id');
  $query->addField('s', 'set_name', 'set_name');
  if (!$id == NULL) {
    $query->condition("s.gc_id", $id, "=");
  }
  $result = $query->execute();
  $ignore = array();
  $options = array();
  $counts = array();
  if (!$result) {
    drupal_set_message(t("No workout sets! You should create some"), 'status');
  }
  else {
    while ($row = $result->fetchAssoc()) {
      if ($exclude == TRUE && $row['gc_id'] == $id) {
        $ignore[$row['set_name']] = 1;
        unset($options[$row['set_name']]);
        continue;
      }
      if (array_key_exists($row['set_name'], $ignore)) {
        continue;
      }
      if (array_key_exists($row['set_name'], $counts)) {
        $counts[$row['set_name']]++;
      }
      else {
        $counts[$row['set_name']] = 1;
      }
    }
  }
  foreach ($counts as $key => $value) {
    $options[$key] = "$key: $value workouts in set";
  }
  return $options;
}

/**
 * Purge activities from the filesystem.
 *
 * All activities are logged. I do this so that users experiencing difficulty
 * can mail me the logs.
 *
 * @param array $activities
 *   An array of activity filesystem locations keyed on activityId
 */
function garmin_connector_purge_activities($activities = array()) {
  if (user_access("manage activities") != TRUE) {
    $message  = t("Sorry, you don't have permission to purge workouts.");
    $message .= t("If you are a site admin, set the 'Manage Activities' permission");
    drupal_set_message($message);
    return;
  }
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  $cat = garmin_connector_get_activities_from_local();
  $by_id = garmin_connector_key_by_id($cat);
  $completed = array();
  global $_garmin_connector_from_drush;
  if ($_garmin_connector_from_drush == 1) {
    $debug = fopen('php://stdout', 'w');
  }
  else {
    $debug_destination = "public://garmin_connector/debug-purge-activities.txt";
    $cpath = drupal_realpath($debug_destination);
    $debug = fopen($cpath, "w");
  }
  fwrite($debug,
      t("@fn These are the activities we need to purge:\n",
      array(
        '@fn' => $fn)
      ) . "$fn " . var_export($activities, TRUE) . "\n");
  foreach ($activities as $id) {
    fwrite($debug,
      t("@fn Working on activity ID (@id)\n",
        array(
          '@fn' => $fn, '@id' => $id)));
    if (!$by_id[$id]) {
      fwrite($debug,
        t("@fn Couldn't find the id (@id). Perhaps you need to rebuild the catalogue. You can either run retrieve_activities or use a drush method.\n",
        array(
          '@fn' => $fn, '@id' => $id)));
    }
    fwrite($debug,
      t("@fn Working on purging @id\n",
        array(
          '@fn' => $fn,
          '@id' => $id)
        )
    );
    $filelist = file_scan_directory('public://garmin_connector',
      "/$id.json|$id.tcx/",
      array('recurse' => TRUE));
    fwrite($debug,
      t("@fn These are the files that I found:\n",
        array(
          '@fn' => $fn)
        ) . var_export($filelist, TRUE) .  "\n");
    foreach ($filelist as $location => $data) {
      fwrite($debug,
        t("Working on the file location: (@location)\n",
          array(
            '@location' => $location)
        ));
      if (file_exists($location)) {
        fwrite($debug,
          t("A file exists at location (@location)\n",
            array(
              '@location' => $location)));
      }
      else {
        $message = t("I cannot purge the file (@file) from the Garmin Connect filesystem because it is not a file.  You can investigate this by checking in the sites/YOURSITE/files/garmin_connect directory",
        array(
          '@file' => $location));
        $completed[$location] = t("FAIL") . ": ($message)";
        fwrite($debug, $message);
        // I need to let them know if they are using a browser.
        drupal_set_message($message, 'warning');
        continue;
      }
      $result = drupal_unlink($location);
      $status = t("FAILED AT drupal_unlink");
      if ($result == TRUE) {
        $message = t("@fn Deleted file @location",
            array(
              '@fn' => $fn,
              '@location' => $location)
        );
        watchdog("garmin-connector",
          "@fn deleted file @location",
          array(
            '@fn' => $fn,
            '@location' => $location,
          )
        );
        fwrite($debug, $message . "\n");
        drupal_set_message($message, 'warning');
        $status = "SUCCESS";
      }
      else {
        watchdog("garmin-connector", "@fn failed to delete the file @location",
        array(
          '@location' => $location)
        );
        $message = t("Failed to delete the file @location",
          array(
            '@location' => $location)
        );
        drupal_set_message($message, 'warning');
        fwrite($debug, t("Deleted the file at location @location\n", array('@location' => $location)));
      }
      $completed[$location] = $status;
    }
  }
  fwrite($debug, t("Done with deleting. Now preparing to rebuild catalogue...") . "\n");
  garmin_connector_rebuild_catalogue($debug);
  fwrite($debug, t("Done rebuilding catalogue.") . "\n");
  return $completed;
}

/**
 * Convert IDs into filesystem locations based on their sport.
 *
 * @param array $ids
 *   An array of workout IDs
 */
function garmin_connector_prepare_purge($ids = array()) {
  $returner = array();
  $workouts = garmin_connector_get_activities_from_local();
  $fixed = garmin_connector_key_by_id($workouts);
  // Locate the one specified.
  foreach ($ids as $value) {
    if (array_key_exists($value, $fixed)) {
      array_push($returner, $value);
    }
    else {
      // This error should go to the browser.
      drupal_set_message(
        "Workout with ID @value not found in the workout catalogue.",
        array(
          '@value' => $value,
        ), 'error');
      continue;
    }
  }
  return $returner;
}

/**
 * Check that the details file and TCX are present for this activity.
 *
 * These files are necessary to cache data for all activities and are
 * built into the system catalog during rebuilds.
 *
 * @param Object $activity
 *   The activity
 *
 * @param bool $debug
 *   A reference to a filehandle for writing debugging messages
 *
 * @param bool $force_details
 *   Force the details file to be written
 *
 * @return int
 *   1 if everything was successful
 */
function garmin_connector_check_activity($activity, $debug, $force_details) {
  $a = $activity->activity;
  $id = $a->activityId;
  $name = $a->activityName;
  $type_parent = garmin_connector_get_type($a);
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  fwrite($debug,
    t("@fn -- Working on an activity id(@id) name(@name) type(@type_parent)\n",
    array(
      '@fn' => $fn,
      '@type_parent' => $type_parent,
      '@id' => $id,
      '@name' => $name)
  ));
  if (!$id) {
    fwrite($debug,
      t("@fn ERROR: Holy cow. This object has no activityId. That can't be. Here's the object:\n@fn object\n",
    array(
      '@fn' => $fn,
      '@object' => var_export($a, TRUE))
    ));
    return 0;
  }
  // Write the summary.
  // We will always write the summaries so that they represent.
  // the latest version.
  // Regardless of anything.
  $time = new DateTime($a->activitySummary->BeginTimestamp->value);
  $time->setTimeZone(
    new DateTimeZone($a->activitySummary->BeginTimestamp->uom));
  garmin_connector_write_json($a, "summary", $type_parent, $time, $debug);
  // This is confusing but I have to roll with it.
  $overview_object = $a;
  if ($type_parent == "swimming" ||
    $type_parent == "cycling" ||
    $type_parent == "running" ||
    $type_parent == "other" ||
    $type_parent == "fitness_equipment") {
  }
  else {
    $ts = array(
      '@id' => $id,
      '@fn' => $fn,
      '@type_parent' => $type_parent);
    drupal_set_message(
      t("@fn -- ERROR Holy cow! Don't know what this activity type is: (@type_parent). Id was (@id)",
      $ts
      ), 'error'
    );
    fwrite($debug,
      t("@fn -- ERROR Holy cow! Don't know what this activity type is: (@type_parent). Id was (@id)",
        $ts
      )
    );
    return 0;
  }
  if (!$type_parent || !$id) {
    fwrite($debug,
      t("@fn -- ERROR Oh dear - there was an activity without a type or id.\n",
      $ts
    )
    );
    fwrite($debug, var_export($a));
    return 0;
  }
  if (!array_key_exists("dir_tcx_" . $type_parent, garmin_connector_get_config_array())) {
    fwrite($debug,
      t("@fn -- ERROR Missing directory configuration for activities of type @type_parent\n",
      $ts
      )
    );
  }
  else {
    garmin_connector_get_and_write_tcx($overview_object, $debug, $force_details);
  }
  $details = garmin_connector_get_and_write_details(
    $overview_object, $debug, $force_details);
  // Now convert the details json into a plottable thing.
  fwrite($debug,
    t("@fn -- Preparing to make json\n",
      array(
        '@fn' => $fn,
      )
    )
  );
  garmin_connector_make_json_from_objects($overview_object, $details, $debug);
  return 1;
}

/**
 * Upload a tcx representing completed strength training to Garmin Connect.
 *
 * The purpose of this is best illustrated with a practical example. You are
 * training for Ironman. You are doing strength workouts that are basically the
 * same, every day. You would like to log those in Garmin Connect. It's really a
 * pain to go there, click "New Activity", set duration X, activity type Y,
 * select the date, the start time, and whatever else Garmin Connect wants.
 * 
 * This method creates a false TCX representing a typical strength workout and
 * uploads it to Garmin Connect. With one click on an ajax-enabled button you've
 * created that workout.
 *
 * The downside to doing this is that, well, it was very very hard to get to
 * work. The Garmin Connect API is poorly documented and does not tolerate any
 * errors at all. Trust me.
 * 
 * This code may seem fragile but I've been using it for about 12 months without
 * any problem.
 *
 * @param string $duration
 *   The Duration
 *
 * @return array
 *   A drupal render array
 *
 * @throws
 *   Errors related to Garmin Connect communication will be reported.
 *
 * @todo
 * In the future, this method will throw proper exceptions.
 */
function garmin_connector_upload_activity($duration) {
  $ch = garmin_connector_initialize();
  $utc = new DateTimeZone("UTC");
  $time = new DateTime('now', $utc);
  // With Saskatchewan, I get tomorrow, 5am so I use UTC.
  $tcx = garmin_connector_get_basic_tcx($time, $duration * 60);
  curl_setopt($ch, CURLOPT_POST, 1);
  $delimiter = '-------------' . uniqid();
  $stringy  = "--" . $delimiter . "\r\n";
  $stringy .= "Content-Disposition: form-data; name=\"responseContentType\"";
  $stringy .= "\r\n";
  $stringy .= "text/html\r\n";
  $stringy .= "--" . $delimiter . "\r\n";
  $stringy .= "Content-Disposition: form-data; name=\"data\"; ";
  $stringy .= "filename=\"csm.tcx\"\r\n";
  $stringy .= "Content-Type: application/octet-stream\r\n";
  $stringy .= "\r\n";
  $stringy .= $tcx;
  $stringy .= "\r\n";
  $stringy .= "--" . $delimiter . "--\r\n";
  curl_setopt($ch, CURLOPT_HEADER, FALSE);
  curl_setopt($ch, CURLOPT_HTTPHEADER, array(
    "Expect:",
      'Content-Length: ' . strlen($stringy),
      'Content-Type: multipart/form-data; boundary=' . $delimiter));
  curl_setopt($ch, CURLOPT_POSTFIELDS, $stringy);
  curl_setopt($ch,
    CURLOPT_URL,
    "http://connect.garmin.com/proxy/upload-service-1.1/json/upload/.tcx");
  $results = curl_exec($ch);
  if (curl_error($ch)) {
    throw new Exception(
      t("There was an error posting new workout data to Garmin Connect. The error was (!error)",
      array(
        '!error' => curl_error($ch),
      )
    ));
  }
  $decoded = json_decode($results);
  $id = NULL;
  $returner = "";
  if ($decoded->detailedImportResult->successes[0]) {
    $id = $decoded->detailedImportResult->successes[0]->internalId;
    $returner .= "<a style='margin-top: 20px; margin-bottom: 20px; ";
    $returner .= "font-size: x-large; display: block;' ";
    $returner .= "href='http://connect.garmin.com/activities/$id'>";
    $returner .= t("Results from creating the new activity in  Garmin Connect");
    $returner .= "</a>";
  }
  else {
    // I wish I could provide more information.
    $returner .= "<p>";
    $returner .= t("Something went wrong when creating the new activity");
    $returner .= "</p>";
  }
  // hook_post_activity allows a user to do post processing when a new strength
  // workout is uploaded to Garmin Connect.
  $all['tcx-uploader'] = array(
    '#id'     => $id,
    '#markup' => $returner,
  );
  return $all;
}

/**
 * Get activities from Garmin Connect.
 * 
 * See the readme for what to do when you've changed a workout
 * in Garmin Connect.
 *
 * @param bool $debug
 *   Debug mode on?
 *
 * @param string $from
 *   The earliest date. Formatted YYYY-MM-DD.
 *
 * @param string $to
 *   The latest date. Formatted YYYY-MM-DD.
 *
 * @return object
 *   An object containing workouts.
 *   The structure of this object is as a decoded activityset
 *   from Garmin Connect's REST interface
 */
function garmin_connector_get_activities_from_gc($debug = NULL,
$from = NULL, $to = NULL) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  if (!$debug) {
    $d1 = "public://garmin_connector/debug-get-activities.txt";
    $f1 = drupal_realpath($d1);
    $debug = fopen($f1, 'w');
  }
  else {
    $debug = fopen('php://stdout', 'w');
  }
  fwrite($debug,
    t("@fn is running...\n",
    array(
      '@fn' => $fn,
    )
  ));
  $ch = garmin_connector_initialize();
  if (!$ch) {
    throw new Exception(t("Garmin connector was unable to get an initialized curl handle and so could not run retreieve_activities. Check the module's issues page a drupal.org to determine what to do next."));
  }
  $refresh_activities = "yes";
  if ($refresh_activities == "no") {
    // The reason you'd do this is for testing purposes.
    $temp = file_directory_temp();
    $activities_json = file_get_contents($temp . "/garmin-connector-activities");
    return $activities_json;
  }
  // Set logical defaults.
  $begin = date(garmin_connector_get_date('two weeks ago'));
  $end = date(garmin_connector_get_date('today'));
  if ($from) {
    $begin = date($from);
  }
  if ($to) {
    $end = date($to);
  }
  $search_opts = array(
    'beginTimestamp'    => $begin,
    'endTimestamp' => $end,
    'start' => 0,
    'limit' => 999,
  );
  if (is_array($search_opts)) {
    // This code left here as a courtesy to anybody who wants to use CURL's..
    // Search_opts. I just could not get it to work.
  }
  // The fun thing is - despite being in the specification
  // activity-search-service does not accept the '>=' or '<=' operator.
  $search_opts_string = "";
  $search_opts_string .= "beginTimestamp>";
  $search_opts_string .= date($begin);
  $search_opts_string .= "&endTimestamp<";
  $search_opts_string .= date($end);
  $search_opts_string .= "&limit=99";
  fwrite($debug,
    t("search_opts_string is @search_opts_string\n",
      array(
        '@fn' => $fn,
        '@search_opts_string' => $search_opts_string)));
  // Call the search.
  global $base_url;
  curl_setopt($ch, CURLOPT_REFERER, $base_url);
  curl_setopt($ch, CURLOPT_POST, FALSE);
  $url = "http://connect.garmin.com/proxy/activity-search-service-1.2/"
    . "json/activities?$search_opts_string";
  curl_setopt($ch, CURLOPT_URL, $url);
  curl_setopt($ch, CURLOPT_HEADER, FALSE);
  fwrite($debug,
    t("@fn Contacting Garmin with this URL: @url\n",
      array(
        '@fn' => $fn,
        '@url' => $url)));
  $activities_json = curl_exec($ch);
  if (curl_error($ch)) {
    watchdog("garmin_connector", "Error (@err) when trying to talk to Garmin Connect to get activities",
      array(
        '@err'  => curl_error($ch),
      )
    );
    fwrite($debug,
      t("@fn There was an error calling Garmin: (@error) URL (!url)\n",
      array(
        '@fn'     => $fn,
        '@error'  => curl_error($ch),
        '!url'    => $url)));
    throw new Exception(t("There was an error calling Garmin Garmin Connect: (@error) URL (!url). If the error is 403 (most common), check your Garmin Connect username and password.\n",
      array(
        '@error'  => curl_error($ch),
        '!url'    => $url)));
  }
  fwrite($debug,
    t("@fn Done calling garmin. This much data was received (@length)\n",
    array(
      '@fn'       => $fn,
      '@length'   => strlen($activities_json))));
  $destination = garmin_connector_get_config('dir_summary') . "/summary.json";
  $filename = drupal_realpath($destination);
  file_put_contents($filename, $activities_json);
  $decoded = json_decode($activities_json);
  return $decoded->results->activities;
}

/**
 * Retrieve TCXs from Garmin Connect (where necessary).
 *
 * TCX is an XML file format for descriing an activity.
 *
 * @param object $object
 *   An object representing a deserialized JSON workout from garmin connect
 *
 * @param bool $force
 *   Force this function to download the TCX even if it already exists
 */
function garmin_connector_get_and_write_tcx($object, $debug, $force = NULL) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  $id = $object->activityId;
  fwrite($debug, t("@fn -- Getting and writing TCX for object with id (@id) Force is (@force)\n",
  array(
    '@fn' => $fn,
    '@id' => $id,
    '@force' => ($force ? "on" : "off"))
  ));
  $ch = garmin_connector_initialize();
  $type_parent = garmin_connector_get_type($object);
  $destination = garmin_connector_get_config(
    'dir_tcx_' . $type_parent) . "/$id.tcx";
  $filename = drupal_realpath($destination);
  if (!$filename) {
    $thing = t("@fn -- the TCX destination (@destination) could not be converted to a filename and so I can't write the file. The activity type was (@type_parent)\n",
    array(
      '@fn' => $fn,
      '@id' => $id,
      '@type_parent' => $type_parent,
      '@destination' => $destination)
    );
    fwrite($debug, $thing);
    drupal_set_message($thing, 'error');
    return;
  }
  $tcx_found = NULL;
  if (file_exists($destination)) {
    fwrite($debug,
      t("@fn -- TCX file for @id already exists.\n",
        array(
          '@fn' => $fn,
          '@id' => $id,
        )
      ));
    $tcx_found = 1;
  }
  if ($tcx_found && !$force) {
    fwrite($debug,
      t("@fn -- TCX found for @id, force off.  Nothing left to do. Bye!\n",
        array(
          '@fn' => $fn,
          '@id' => $id,
        )
      ));
    return;
  }
  fwrite($debug,
    t("@fn -- Getting a TCX file for activity with id @id and type of @type_parent\n",
      array(
        '@fn' => $fn,
        '@id' => $id,
        '@type_parent' => $type_parent)
    ));
  $url = "https://connect.garmin.com/proxy/activity-service-1.2/tcx/";
  $url .= "activity/" . $id . "?full=true";
  // Decide if we already have the correct tcx.
  curl_setopt($ch, CURLOPT_HEADER, FALSE);
  curl_setopt($ch, CURLOPT_URL, $url);
  $result = curl_exec($ch);
  if ($filename) {
    file_put_contents($filename, $result);
    $time = new DateTime($object->activitySummary->BeginTimestamp->value);
    $time->setTimeZone(new DateTimeZone(
      $object->activitySummary->BeginTimestamp->uom));
    touch($filename, $time->getTimestamp());
  }
  else {
    fwrite($debug,
      t("@fn -- Could not write the TCX because there was no filename.\n",
      array(
        '@fn' => $fn)
    ));
  }
}

/**
 * If necessary, initialize all that's necessary to call GConnect via REST.
 *
 * Don't bother doing this if there is no Garmin Connect username set.
 *
 * @param filehandle $debug
 *   A filehandle to write debugging information
 *
 * @return object
 *   An initialized, logged-in CURL handle.
 */
function garmin_connector_initialize($debug = FALSE) {
  $credentials = garmin_connector_get_credentials();
  if (is_array($credentials)) {
    // Do nothing. This was only here to throw early if no username is set.
  }
  global $_garmin_connector_is_initialized;
  global $_garmin_connector_ch;
  if ($_garmin_connector_is_initialized == TRUE) {
    $debug && fwrite($debug, t("Garmin Connector is already initialized. Returning.\n"));
    return $_garmin_connector_ch;
  }
  $ch = curl_init();
  $debug && fwrite($debug, t("Garmin Connector is preparing to setup curl.\n"));
  $ch = garmin_connector_gc_setup($ch, $debug);
  // The result of this will be a logged in handle.
  $ch = garmin_connector_gc_login($ch);
  $_garmin_connector_is_initialized = TRUE;
  $_garmin_connector_ch = $ch;
  return $ch;
}

/**
 * Prepare CURL options necessary to connect to Garmin Connect.
 *
 * @param object $ch
 *   A CURL handle.
 *
 * @param filehandle $debug
 *   A filehandle to write debug information.
 *
 * @return object
 *   An curl handle with the options set to allow connections to Garmin Connect
 */
function garmin_connector_gc_setup($ch, $debug = NULL) {
  curl_setopt($ch, CURLOPT_POST, TRUE);
  curl_setopt($ch, CURLOPT_FAILONERROR, TRUE);
  curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
  curl_setopt($ch, CURLOPT_HEADER, TRUE);
  $temp = file_directory_temp();
  global $user;
  if (user_is_logged_in()) {
    curl_setopt($ch, CURLOPT_COOKIEJAR, $temp . "/garmin-cookies-" . $user->name);
  }
  else {
    curl_setopt($ch, CURLOPT_COOKIEJAR, $temp . "/garmin-cookies");
  }
  curl_setopt($ch, CURLOPT_FOLLOWLOCATION, TRUE);
  return $ch;
}

/**
 * Verify that the user has set a Garmin Connect username and password.
 *
 * @throws
 *   An exception if the username or password are blank.
 *
 * @return array
 *   In single user mode (if you are reading this then Garmin Connector only
 *   supports one user per Drupal install) return an associative array with
 *   the keys 'username' and 'password'.
 *   
 * @todo
 *   To support multiple users per Drupal install, I'd store these in a table
 *   rather than as a variable, take the username as an argument (or get it
 *   from logged in user), and return the password.
 */
function garmin_connector_get_credentials() {
  $username = variable_get("garmin_connector_username", "unknown");
  $password = variable_get("garmin_connector_pw", "blank");
  if ($username == "unknown") {
    throw new Exception(t("You did not set your username in Garmin Connect's !settings page",
    array(
      '!settings' => url('admin/config/garmin_connector/garmin_connector'),
    )));
  }
  return array(
    'username' => $username,
    'password' => $password,
  );
}

/**
 * Check logged in status without contacting Garmin Connect to verify.
 *
 * This produces a check that may not be reliable and should not be trusted.
 * It's for estimation purposes only and was specifically written for Drupal
 * system status pages where the delay due to an *actual* lookup was
 * unacceptable.
 *
 * return array
 *   An estimation (with explanation) of the Garmin Connect login status.
 *   'value'       - A value usable by hook_requirements
 *   'severity'    - A severity constant defined by hook_requirements
 *   'description' - A description usable by hook_requirments
 */
function garmin_connector_estimate_login_status() {
  $last_checked = variable_get('garmin_connector_retrieve_last_run', 'never');
  if ($last_checked == "never") {
    return array(
      'title'    => t("Garmin Connector: Able to log into Garmin Connect?"),
      'severity' => REQUIREMENT_WARNING,
      'value'    => t("Not ok"),
      'description'  => t("Garmin connector has never successfully contacted Garmin Connect. Perhaps you need to set your Garmin Connect username and password !here or retrieve activities !there.",
        array(
          '!here' => "<a href='" . url("admin/config/garmin_connector/garmin_connector") . "'>" . t("here") . "</a>",
          '!there' => "<a href='" . url("garmin-connector/activity-manager.html") . "'>" . t("here") . "</a>",
        )
      ),
    );
  }
  $last_date = new DateTime($last_checked);
  $now = new DateTime();
  $interval = $now->diff($last_date);
  $hours = $interval->format('h');
  // Three day threshold.
  if ($hours < 72) {
    // This is a massive assumption, but let's go with it because we have to...
    // Save performance.
    return array(
      'title'    => t("Garmin Connector: Able to log into Garmin Connect?"),
      'value'    => "ok",
      'severity' => REQUIREMENT_OK,
      'description' => t("Garmin Connector talked to Garmin Connect less than 72 hours ago on (@here). You logged in successfully at that time. Good!",
      array(
        '@here' => $last_date->format("Y:m:d H:i:s"),
      )
      ),
    );
  }
  else {
    return array(
      'title'    => t("Garmin Connector: Able to log into Garmin Connect?"),
      'value'    => t("ok"),
      'severity' => REQUIREMENT_OK,
      'description' => t("Garmin Connector talked to Garmin Connect earlier than 72 hours ago. It was on (@here) but was successful so that seems good.",
      array(
        'when' => $last_date->format("Y:m:d H:i:s"),
      )
      ),
    );
  }
}


/**
 * Verify that the user is logged into Garmin Connect.
 *
 * @param object $ch
 *   A CURL handle
 *
 * @param filehandle $debug
 *   A filehandle to write debugging information
 *
 * @return bool
 *   A boolean indicating status. TRUE is logged in, FALSE is not logged in
 *
 * @throws Exception
 *   If there are problems logging in, errors will be reported.
 */
function garmin_connector_gc_is_logged_in($ch, $debug = NULL) {
  global $_garmin_connector_is_logged_in;
  if (isset($_garmin_connector_is_logged_in) && $_garmin_connector_is_logged_in == TRUE) {
    return TRUE;
  }
  // Get the main page.
  $debug = "true";
  if ($debug) {
    // curl_setopt($ch, CURLOPT_VERBOSE, TRUE);
  }
  curl_setopt($ch, CURLOPT_POST, FALSE);
  curl_setopt($ch, CURLOPT_URL, "http://connect.garmin.com");
  $result = curl_exec($ch);
  $username = variable_get("garmin_connector_username");
  if (curl_error($ch)) {
    // This should never happen. It's quite catastrophic.
    throw new Exception(
      t("There was an error asking Garmin Connect if we are logged in. The error was (!err)",
      array(
        '!err' => curl_error($ch),
      )));
  }
  // There's a place where the username of a logged in user is displayed.
  $now = new DateTime();
  if (strpos($result, $username) > 0) {
    $_garmin_connector_is_logged_in = TRUE;
    return TRUE;
  }
  else {
    // I can't throw an exception here. Let's log some information.
    if (!isset($debug)) {
      $d1 = "public://garmin_connector/debug-dashboard.txt";
      $f1 = drupal_realpath($d1);
      $debug = fopen($f1, 'w');
    }
    fwrite($debug, t("There is a problem logging into Garmin Connect. Here is some detail\n"));
    fwrite($debug, t("Problem occurred at !time\n",
      array(
        '!time' => $now->format("Y:m:d H:i:s"),
      )
    ));
    $callers = debug_backtrace();
    $caller = var_export(array_shift($callers), TRUE);
    fwrite($debug, t("The caller:\n"));
    fwrite($debug, $caller . "\n");
    fwrite($debug, t("An export of curl_getinfo:\n"));
    $info = var_export(curl_getinfo($ch), TRUE);
    fwrite($debug, $info . "\n");
    fwrite($debug, t("The HTML of the resulting page:\n"));
    fwrite($debug, $result);
    watchdog("garmin_connector",
      "Garmin Connector was unable to find a username (@username) in the Garmin Connect dashboard. It's possible that you are not logged in or that the credentials you specified to connect to Garmin Connect were not correct. The length of the response I got from Garmin Connect was (!length). This should be a bit more than 10,000. If it isn't perhaps Garmin Connect is not available. The caller was @caller. Here is more information !info",
      array(
        '@username' => $username,
        '!length' => strlen($result),
        '@caller' => $caller,
        '!info' => var_export($info, TRUE),)
     );
  }
  return FALSE;
}

/**
 * Log the user into Garmin Connect.
 *
 * This was *super* tricky and I spent a long time trying to get it
 * to work with PHP/CURL.
 *
 * Basically, a cookie is set with the session ID of a logged-in session
 * That's why there's a bunch of messing around with cookie files
 *
 * Note that you can't just POST to the login screen - you need to get an empty
 * cookie first.
 *
 * @param object $ch
 *   A CURL handle
 *
 * @param filehandle $debug
 *   A filehandle to print informative messages
 *
 * @return object
 *   A CURL handle that's logged into Garmin Connect
 *
 * @throws Exception
 *   - Errors related to having no username or password set will be reported.
 *   - Errors related to network communication will be reported.
 *
 * @throws
 *   Exception when not logged in.
 */
function garmin_connector_gc_login($ch, $debug = NULL) {
  $credentials = garmin_connector_get_credentials();
  // First of all, are we already logged in?
  if (garmin_connector_gc_is_logged_in($ch)) {
    return $ch;
  }
  global $_garmin_connector_from_drush;
  if (!user_is_logged_in() && !$_garmin_connector_from_drush) {
    throw new Exception(t("You must be logged in to this website to use Garmin Connector. Log in !here",
    array(
      '!here' => "<a href='" . url("user/login") . "'>" . t("here") . "</a>",
    )));
  }
  $url_gc_login    = 'https://connect.garmin.com/signin';
  curl_setopt($ch, CURLOPT_URL, $url_gc_login);
  // The reason why we ignore the output is because we care only about the..
  // Cookie that Garmin Connect sets.
  $result1 = curl_exec($ch);
  if (!$result1 || curl_error($ch)) {
    throw new Exception(t("An unknown error occurred while trying to set the initial cookie for Garmin Connect. The error was !error.",
      array(
        '!error' => curl_error($ch),
      )
    ));
  }
  curl_setopt($ch, CURLOPT_POST, TRUE);
  $loginopts = array(
    'login'                   => 'login',
    'login:signInButton'      => 'Sign In',
    // This stopped working in December 2013 and I had to change j_id1 to j_id2.
    // I really have no idea what that means but I included that information..
    // Here in case something happens in the future.
    'javax.faces.ViewState'   => 'j_id2',
    'login:loginUsernameField' => $credentials['username'],
    'login:password'          => $credentials['password'],
    'login:rememberMe'        => 'on',
  );
  curl_setopt($ch, CURLOPT_POST, TRUE);
  curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($loginopts));
  curl_setopt($ch, CURLOPT_URL, $url_gc_login);
  curl_setopt($ch, CURLOPT_REFERER, $url_gc_login);
  $result2 = curl_exec($ch);
  if ($result2 === FALSE || curl_error($ch)) {
    throw new Exception(
      t("There was an error logging into Garmin Connect. The error was (!err)",
      array(
        '!err' => curl_error($ch),
      )));
  }
  return $ch;
}

/**
 * Refresh an activity by downloading necessary stuff from Garmin Connect.
 *
 * This initially did not seem to be possible. If you get the details for an
 * activity via activity-service-1.2, you get a different object than if you
 * save the stream of objects received via activity-search-service.
 *
 * You cannot use activity-search-service to get the activity because although
 * activityId is listed as a searchable field you get a:
 * "HTTP Status 400 - Searching the field "Activity Id" is not currently
 * supported." error.
 *
 * After pondering this for a while, I decided to do this:
 * Using activity-service, get the activity and find out when it was performed
 * and then search for activities (using activity-search-service) with the exact
 * same start time.
 *
 * This is a horrible kludge but Garmin simply ignores any requests for help on
 * their middleware.
 *
 * @param int $id
 *   The activity ID
 *
 * @throws Exception
 *   - An exception if the id cannot be found
 *   - An exception if Garmin Connect could not be contacted
 *   - An exception if Garmin Connect did not return data
 */
function garmin_connector_refresh_activity($id) {
  $fn = __FUNCTION__;
  // Output to stdout when using drush, else to a file.
  global $_garmin_connector_from_drush;
  if (isset($_garmin_connector_from_drush)) {
    $debug = fopen('php://stdout', 'w');
  }
  else {
    $d1 = "public://garmin_connector/debug-refresh-activity.txt";
    $f1 = drupal_realpath($d1);
    $debug = fopen($f1, 'w');
  }
  $ch = garmin_connector_initialize();
  fwrite($debug,
    t("@fn -- Getting detail and overview from Garmin Connect...\n",
    array(
      '@fn' => $fn,
    )
  ));
  curl_setopt($ch, CURLOPT_HEADER, FALSE);
  curl_setopt($ch, CURLOPT_POST, FALSE);
  // The object that activity-service-1.2 delivers is not the object we want
  // but we need the dates for the activity.
  $overview_url = "http://connect.garmin.com/proxy/activity-service-1.2/";
  $overview_url .= "json/activity/" . $id;
  curl_setopt($ch, CURLOPT_URL, $overview_url);
  fwrite($debug,
    t("@fn -- Getting overview with this url: (@overview_url)\n",
    array(
      '@fn' => $fn,
      '@overview_url' => $overview_url,
    )
    )
  );
  $overview_result = curl_exec($ch);
  if (curl_error($ch)) {
    // This can (and will) fail. Having this info will help in debugging.
    $message = t("@fn -- Curl call to retrieve overview failed with this error: (@curl_error). This is a fatal error. The activity cannot be refreshed.\n",
      array(
        '@fn' => $fn,
        '@curl_error' => curl_error($ch),
      ));
    fwrite($debug, $message);
    throw new Exception($message);
  }
  fwrite($debug, t("@fn -- Got overview result of length (@length)\n",
    array(
      '@fn' => $fn,
      '@length' => strlen($overview_result),
    )
  ));
  $overview_object = json_decode($overview_result);
  // Get the start time of the activity. This must exactly match GC's date.
  $millis = $overview_object->activity->activitySummaryBeginTimestamp->millis / 1000;
  $date = DateTime::createFromFormat('U', $millis);
  $date->setTimeZone(
    new DateTimeZone($overview_object->activity->activityTimeZone->key)
  );
  $search_url = "http://connect.garmin.com/proxy/activity-search-service-1.2/"
    . "json/activities?beginTimestamp="
    . $date->format(DATE_ATOM);
  fwrite($debug, t("@fn -- Now searching with this url (@search_url)...\n",
    array(
      '@fn' => $fn,
      '@search_url' => $search_url)
  ));
  curl_setopt($ch, CURLOPT_URL, $search_url);
  $search_json = curl_exec($ch);
  if (curl_error($ch)) {
    $message = t("@fn -- Curl call to Garmin connector with search url !search_url failed with this error: (@curl_error)\n",
      array(
        '!search_url' => $search_url,
        '@fn' => $fn,
        '@curl_error' => curl_error($ch),
      )
    );
    fwrite($debug, $message);
    throw new Exception($message);
  }
  fwrite($debug,
    t("@fn -- Got search result of length (@length)\n",
      array(
        '@fn' => $fn,
        '@length' => strlen($search_json),
      )
  ));
  $search_object = json_decode($search_json);
  $activities = $search_object->results->activities;
  if (count($search_object->results->activities) != 1) {
    $message = t("@fn -- When searching for your activity, I found more than or less than one activity. This is horrible and fatal for @fn\n",
      array(
        '@fn' => $fn,
      )
    );
    fwrite($debug, $message);
    throw new Exception($message);
  }
  $myactivity = $activities[0];
  // REMEMBER - $overview_object WILL NOT WORK HERE.
  garmin_connector_check_activity($myactivity, $debug, TRUE);
  fwrite($debug, t("@fn is asking for json to be written\n",
    array(
      '@fn' => $fn,
    ))
  );
  // garmin_connector_make_json_from_objects($myactivity, $details, $debug);
  fwrite($debug, t("@fn is finished\n",
    array(
      '@fn' => $fn,
    ))
  );
  garmin_connector_rebuild_catalogue();
}

/**
 * Go to GC to get a json file.
 *
 * If the json file representing details on an activity is absent, 
 * go to Garmin Connect and get it
 *
 * @param Object $overview_object
 *   An object representing the activity. This is an object expected to be
 *   derived from the json_decode-ing of an activity from Garmin Connect
 *
 * @param bool $force
 *   Should the system be forced to go and download the details, even if
 *   they exist on the filesystem
 *
 * @return Object
 *   A details object as defined by Garmin Connect
 */
function garmin_connector_get_and_write_details($overview_object, $debug,
$force = NULL) {
  $ch = garmin_connector_initialize();
  $id = $overview_object->activityId;
  $type_parent = garmin_connector_get_type($overview_object);
  $name = $overview_object->activityName;
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  fwrite($debug,
    t("@fn -- Working with an activity parent(@type_parent) name(@name) id(@id)\n",
    array(
      '@fn'           => $fn,
      '@type_parent'  => $type_parent,
      '@id'           => $id,
      '@name'         => $name)
  ));
  $destination2 = garmin_connector_get_config('dir_details_' . $type_parent) . "/$id.json";
  $filename = drupal_realpath($destination2);
  $force_download = 1;
  if ($force_download != 1 && is_file($filename) && !$force && !filesize($filename) == 0) {
    fwrite($debug,
      t("@fn -- A file (@filename) for @id exists so I'm skipping the download)\n",
      array(
        '@fn'       => $fn,
        '@id'       => $id,
        '@filename' => $filename)));
    $json = file_get_contents($filename);
    $details = json_decode($json);
  }
  else {
    if (is_file($filename) && filesize($filename) == 0) {
      // It would be good to fail with an exception here. In the future, add...
      // Exceptions here.
      fwrite($debug,
        t("@fn Since the filesize of @filename is zero, something must have gone wrong when trying to retrieve it before.",
      array(
        '@fn' => $fn,
        '@filename' => $filename)));
    }
    curl_setopt($ch, CURLOPT_HEADER, FALSE);
    curl_setopt($ch,
      CURLOPT_URL,
      "http://connect.garmin.com/proxy/activity-service-1.2/json/activityDetails/" . $id);
    $json = curl_exec($ch);
    $strlen = strlen($json);
    // For some reason Garmin uses this long string in Details.
    // Get rid of it.
    $json = preg_replace("/com.garmin.activity.details.json.ActivityDetails/",
      "activity",
      $json);
    fwrite($debug,
      t("@fn -- Got activityDetails data from gc. Length of data was (@len)\n",
      array(
        '@fn' => $fn,
        '@len' => $strlen)));
    if (strlen($json) == 0) {
      fwrite($debug,
      t("@fn -- No response came from garmin there. Here's the info @info",
      array(
        '@fn' => $fn,
        '@info' => var_export(curl_getinfo($ch), TRUE))));
    }
    $details = json_decode($json);
    $time = new DateTime(
      $overview_object->activitySummary->BeginTimestamp->value);
    $time->setTimeZone(new DateTimeZone(
      $overview_object->activitySummary->BeginTimestamp->uom));
    $details->summary = $overview_object;
    garmin_connector_write_json(
      $details->activity,
      "details",
      $type_parent,
      $time,
      $debug);
  }
  $details->filename = $filename;
  return $details;
}

/**
 * Get a list of sports.
 *
 * @return array
 *   An array of sports
 */
function garmin_connector_get_sports() {
  return array("swimming", "cycling", "running", "other");
}


/**
 * Get the type of the object.
 *
 * @param object $object
 *   The object to analyse
 *
 * @return string
 *   The type
 */
function garmin_connector_get_type($object) {
  $type_parent = $object->activityType->parent->key;
  if ($type_parent == "all") {
    return $object->activityType->key;
  }
  return $type_parent;
}

/**
 * Write json.
 *
 * All data related to activities are stored as JSON. The structure of the JSON
 * files are of the same format that Garmin Connect's API generates.
 *
 * @param object $object
 *   A garmin connect object
 *
 * @param string $type
 *   "details" or "summary"
 *
 * @param string $activitytype
 *   running, cycling, other and so forth. 
 *   This has to be included because the details json doesn't have it
 *
 * @param string $time
 *   The date of the activity as found here:
 *   $object->activitySummary->BeginTimestamp->value
 *   This has to be included because the details json doesn't have it
 *
 * @param filehandle $debug
 *   A filehandle to write debug messages
 */
function garmin_connector_write_json($object, $type, $activitytype, $time,
$debug) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  $id = $object->activityId;
  // Try harder.
  if (!isset($id)) {
    $id = $object->{'com.garmin.activity.details.json.ActivityDetails'}->activityId;
  }
  if (!isset($id)) {
    $message = t("@fn -- there is no id here.",
      array(
        '@fn' => $fn,
      )
    );
    fwrite($debug, $message);
    // This happens rarely, if ever.
  }
  $key = 'dir_' . $type . '_' . $activitytype;
  if (!array_key_exists($key, garmin_connector_get_config_array())) {
    // This is a bad error but lets not drop dead.
    fwrite($debug,
      t("I cannot find the key (@key) in the configuration and so don't 
      know where to write json. id was (@id)",
      array(
        '@key' => $key,
        '@id' => $id,
      )
    ));
    return;
  }
  $destination = garmin_connector_get_config($key) . "/$id.json";
  $filename = drupal_realpath($destination);
  if (!$filename) {
    $message
      = t("@fn -- No filename could be generated for the destination @destination.",
      array(
        '@fn' => $fn,
        '@destination' => $destination,
      )
    );
    fwrite($debug, "$message\n");
    // It's a good idea to let the user from a browser know that there was...
    // A problem. Silent failure is bad.
    drupal_set_message($message, 'error');
    return;
  }
  $decoded = garmin_connector_get_json_string($object);
  fwrite($debug,
    "$fn -- Writing $type json for ($id) to $filename\n");
  file_put_contents($filename, $decoded);
  // It is very important that the file be timestamped properly.
  touch($filename, $time->getTimestamp());
}

/**
 * Get a pretty formatted JSON string.
 *
 * This is really for readability.
 * I used to pretty print but that isn't available in < php 5.4.
 *
 * @param object $object 
 *   The object to JSON encode
 *
 * @return string
 *   A JSON encoded object
 */
function garmin_connector_get_json_string($object) {
  if (phpversion() >= 5.4) {
    return json_encode($object, JSON_PRETTY_PRINT);
  }
  /* JSON_PRETTY_PRINT not available. */
  return json_encode($object);
}

/**
 * Get the latest activity from the filesystem.
 *
 * Note that this requires that garmin_connector_retreieve_activities 
 * ran successfully and that the system catalogue was regenerated.
 *
 * @param string $sport_regex
 *   A regex for the sport to get
 *
 * @param bool $use_ignore_regex
 *   Use the regex
 *
 * @return object
 *   An object that looks like a json_decode-d activity from Garmin Connect
 */
function garmin_connector_get_latest($sport_regex = NULL, $use_ignore_regex = FALSE) {
  $ignore_regex = variable_get(
    'garmin_connector_key_workout_block_ignore_regex',
    "foo");
  $workouts = garmin_connector_get_activities_from_local();
  if ($sport_regex == NULL) {
    // No need to do much here.
    // Get the system catalog by date.
    $latest = $workouts[0];
    return $latest;
  }
  foreach ($workouts as $workout) {
    $type_parent = garmin_connector_get_type($workout);
    $name = $workout->activityName;
    if (preg_match("/$sport_regex/", $type_parent)) {
      if ($use_ignore_regex == FALSE) {
        return $workout;
      }
      $matches = preg_match("/$ignore_regex/i", $name);
      if ($matches == 0) {
        return $workout;
      }
    }
  }
}

/**
 * Write JSON to the filesystem.
 *
 * Retreives measurements from Garmin Connect where necessary.
 *
 * @param object $overview
 *   An overview object
 *
 * @param object $details
 *   The details object
 */
function garmin_connector_make_json_from_objects($overview, $details, $debug) {
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  if (!is_object($overview)) {
    fwrite($debug,
      t("ERROR: overview is not an object") .
      var_export($overview, TRUE) . "\n");
    return;
  }
  elseif (!$overview->activityId) {
    fwrite($debug,
    t("ERROR: overview has no activityId") .
    var_export($overview, TRUE) . "\n");
    return;
  }
  if (!is_object($details)) {
    $callers = debug_backtrace();
    $caller = var_export(array_shift($callers), TRUE);
    fwrite($debug,
      t("@fn ERROR: details is not an object. Caller was !caller.",
      array(
        '@fn' => $fn,
        '!caller' => $caller,
      )) .
      var_export($details, TRUE) . "\n");
    return;
  }
  if (!is_object($details->activity)) {
    $callers = debug_backtrace();
    $caller = var_export(array_shift($callers), TRUE);
    fwrite($debug,
      t("@fn This is a new one. details->activity is not an object. Why not??  Caller (!caller)\n!dump\n",
      array(
        '@fn' => $fn,
        '!caller' => $caller,
        '!dump' => var_export($details, TRUE),
      )
    ));
    return;
  }
  $points = array();
  if (!$details) {
    fwrite($debug,
      t("@fn There is no activity for activity with id @id in the details\n",
      array(
        '@fn' => $fn,
        '@id' => $overview->activityId,
      )
    ));
  }
  if (!property_exists($details->activity, 'measurements')) {
    fwrite($debug,
      t("@fn -- There are no measurements in the details for id (@id)\n",
      array(
        '@fn' => $fn,
        '@id' => $overview->activityId))
      );
  }
  else {
    fwrite($debug,
      t("@fn -- Retrieving measurements....\n",
        array(
          '@fn' => $fn)
    ));
    $points = garmin_connector_get_measurements($details, $debug);
    fwrite($debug,
      t("@fn -- Done getting measurements.\n",
      array(
        '@fn' => $fn,
      )
      )
    );
  }
  $id = $overview->activityId;
  if (!$id) {
    fwrite($debug,
      t("@fn This activity has no activity id.\n",
      array(
        '@fn' => $fn,
      )
    ));
    return;
  }
  else {
    $destination_json = garmin_connector_get_config('dir_plot_data') . "/" . $id . ".json";
    $destination_filename = drupal_realpath($destination_json);
    fwrite($debug, t("@fn -- Destination json filename is @destination_json\n",
      array(
        '@fn' => $fn,
        '@destination_json' => $destination_json)));
    $points['summary'] = $overview;
    file_put_contents($destination_filename, garmin_connector_get_json_string($points));
  }
}

/**
 * Refresh activities from Garmin Connector.
 *
 * Connect to Garmin Connect and retrieve any new activities. Pre-parse them to
 * make any systems relying on Garmin Connector work faster
 *
 * Fun side-effect: Your TCX files are backed up!
 *
 * Connecting to Garmin Connect is a very challenging process that often
 * requires some amount of troubleshooting. The debug information is critical in
 * solving problems.
 * 
 * I use drush a LOT when doing this troubleshooting and that's why there's code
 * to output to stdout. Checkout garmin_connector.drush.inc or ask drush help
 * for more information on my functions for running retreieve-activities 
 * on the commandline.
 *
 * Exceptions are also critical - they allow the correct level of code to
 * intercept any problems. I've taken great effort to throw logical and clear
 * exceptions that rarely make it to the browser if the browser is the parent
 * caller.
 *
 * @throws Exception
 *   If there's a problem rebuilding the catalogue an explanation will be
 *   provided.
 *
 * @see garmin_connector_cron
 */
function garmin_connector_retrieve_activities() {
  module_load_include('inc', 'garmin_connector', 'garmin_connector.catalogue');
  $fn = preg_replace("/garmin_connector_/", "", __FUNCTION__);
  // Behave differently when called from drush.
  global $_garmin_connector_from_drush;
  // Set up output.
  if (isset($_garmin_connector_from_drush)) {
    $debug = fopen('php://stdout', 'w');
  }
  else {
    $d1 = "public://garmin_connector/debug-retrieve-activities.txt";
    $f1 = drupal_realpath($d1);
    $debug = fopen($f1, 'w');
  }
  fwrite($debug, t("@fn is running and getting activities from Garmin Connect\n",
    array('@fn' => $fn)));
  try {
    // These time frames are hard coded now but will be moved to the
    // administrative interface later.
    $activities = garmin_connector_get_activities_from_gc($debug,
      garmin_connector_get_date('one month ago'),
      garmin_connector_get_date('today')
    );
    fwrite($debug,
    t("@fn is done getting activities.\n",
    array('@fn' => $fn)));
    $now = new DateTime();
    variable_set('garmin_connector_retrieve_last_run', $now->format("Y:m:d H:i:s"));
  }
  catch (Exception $f) {
    fwrite($debug,
      t("@fn failed to get activities. Error was !error",
      array(
        '@fn' => $fn,
        '!error' => $f->getMessage())));
    throw $f;
  }
  // Garmin_connector_retrieve_activities() does more than just retrieve..
  // Activities from Garmin Connect so we can't throw an exception when..
  // Retrieval fails.
  if (isset($activities)) {
    foreach ($activities as $activity) {
      $was_downloaded = garmin_connector_check_activity(
          $activity,
          $debug,
          NULL);
      if (!isset($was_downloaded)) {
      }
    }
  }
  // Any time there are changes to the activities on the filesystem we need...
  // To rebuild the catalogue.
  garmin_connector_rebuild_catalogue();
  $time = new DateTime();
  variable_set("garmin_connector_retrieve_last_run", $time->format("Y:m:d H:i:s"));
  if (isset($f)) {
    throw new Exception(t("Although the system catalogue was rebuilt successfully, there was a problem retrieving activities. The problem was (!problem).",
    array(
      '!problem' => $f->getMessage(),
    )));
  }
}

/**
 * Create a fake TCX for uploading to Garmin Connect.
 *
 * Chad uses this as a convenience method to add a new Activity 
 * to Garmin Connect.
 *
 * @param DateTime $starttime
 *   The start time of the activity
 *
 * @param int $time_seconds
 *   The duration of the activity (in seconds)
 *
 * @return string
 *   A fake TCX
 */
function garmin_connector_get_basic_tcx($starttime, $time_seconds) {
  $returner = "";
  $returner .= "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
  $returner .= "<TrainingCenterDatabase ";
  $returner .= "xsi:schemaLocation=\"http://www.garmin.com/xmlschemas/TrainingCenterDatabase/v2 http://www.garmin.com/xmlschemas/TrainingCenterDatabasev2.xsd\" ";
  $returner .= "xmlns:ns5=\"http://www.garmin.com/xmlschemas/ActivityGoals/v1\" xmlns:ns3=\"http://www.garmin.com/xmlschemas/ActivityExtension/v2\" ";
  $returner .= "xmlns:ns2=\"http://www.garmin.com/xmlschemas/UserProfile/v2\" xmlns=\"http://www.garmin.com/xmlschemas/TrainingCenterDatabase/v2\" ";
  $returner .= "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:ns4=\"http://www.garmin.com/xmlschemas/ProfileExtension/v1\">\n";
  $returner .= "  <Activities>\n";
  $returner .= "    <Activity Sport=\"Other\">\n";
  $returner .= "   <Id>" . $starttime->format("Y-m-d\TH:i:s") . "</Id>\n";
  $returner .= "   <Lap StartTime=\"" . $starttime->format("Y-m-d\TH:i:s\Z") . "\">\n";
  $returner .= "     <TotalTimeSeconds>" . $time_seconds . ".0</TotalTimeSeconds>\n";
  $returner .= "     <DistanceMeters>0.0</DistanceMeters>\n";
  // $returner .= "     <Calories>0</Calories>\n";
  // $returner .= "     <Intensity>Active</Intensity>\n";
  $returner .= "     <TriggerMethod>Manual</TriggerMethod>\n";
  $returner .= "   </Lap>\n";
  $returner .= "   <Creator xsi:type=\"Device_t\">\n";
  $returner .= "     <Name>Garmin Connect</Name>\n";
  $returner .= "     <UnitId>0</UnitId>\n";
  $returner .= "     <ProductID>1</ProductID>\n";
  $returner .= "     <Version>\n";
  $returner .= "    <VersionMajor>12</VersionMajor>\n";
  $returner .= "    <VersionMinor>5</VersionMinor>\n";
  $returner .= "    <BuildMajor>15</BuildMajor>\n";
  $returner .= "    <BuildMinor>0</BuildMinor>\n";
  $returner .= "     </Version>\n";
  $returner .= "   </Creator>\n";
  $returner .= "    </Activity>\n";
  $returner .= "  </Activities>\n";
  $returner .= "  <Author xsi:type=\"Application_t\">\n";
  $returner .= "    <Name>Garmin Connect API</Name>\n";
  $returner .= "    <Build>\n";
  $returner .= "   <Version>\n";
  $returner .= "     <VersionMajor>12</VersionMajor>\n";
  $returner .= "     <VersionMinor>5</VersionMinor>\n";
  $returner .= "     <BuildMajor>15</BuildMajor>\n";
  $returner .= "     <BuildMinor>15</BuildMinor>\n";
  $returner .= "   </Version>\n";
  $returner .= "    </Build>\n";
  $returner .= "    <LangID>en</LangID>\n";
  $returner .= "    <PartNumber>006-D2449-00</PartNumber>\n";
  $returner .= "  </Author>\n";
  $returner .= "</TrainingCenterDatabase>\n";
  return $returner;
}
